// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// RadmonFileMaker.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Creates a spectrum using evselect
//
#include <map>
#include <Tracer.h>
#include <metatask.h>
#include <errstr.h>
#include <param.h>
#include <Dal.h>
#include <DalVec.h>
#include <SasDalUtils.h>
#include "TimeTags.h"
#include "RadmonFileMaker.h"



RadmonFileMaker::RadmonFileMaker(std::vector<std::string> radmonVec, std::string &outputfile)
{
  PROC;

  // Create the new file
     DataSet *outSet = dataSetServer->open(outputfile, Dal::Create);

  // Split the files into ECX and ESX
     std::vector<std::string> ecxfiles, esxfiles, allfiles;
     for (unsigned i=0; i<radmonVec.size(); i++) {
    	 DataSet *set = dataSetServer->open(radmonVec[i],DataSet::Read);
    	 if (set->hasBlock("RMECX1") )
    	 {
    		 ecxfiles.push_back(radmonVec[i]);
            allfiles.push_back(radmonVec[i]);
    	 }
    	 else if (set->hasBlock("RMESX1") )
    	 {
            esxfiles.push_back(radmonVec[i]);
            allfiles.push_back(radmonVec[i]);
    	 }
    	 dataSetServer->close(set);
     }

  // Handle the radmon count rate data
     if (ecxfiles.size()>0)
        writeCountRateBlock(outSet, ecxfiles);

  // Handle the radmon spectral data
     if (booleanParameter("includespectra"))
    	 if (esxfiles.size()>0)
    		 writeSpectralBlock(outSet, esxfiles);

     modifyOutputKeywords(outSet,allfiles);

  // Close output set
     dataSetServer->close(outSet);

}


/* Read DATA-END keyword and modify the outuput file accordingly */
void RadmonFileMaker::modifyOutputKeywords(DataSet *outSet,  std::vector<std::string> allfiles)
{

  errstr << "Setting DATE-END..." << Sas::message(Msg::App, Msg::Always);
  STime dateend, tmpdate;
  for (unsigned int i=0; i<allfiles.size(); i++) {
    DataSet *set = dataSetServer->open(allfiles[i],DataSet::Read);

    if (set->hasBlock("RMECX1"))
    {
    	tmpdate = set->table("RMECX1")->attribute("DATE-END")->asString();
    	if (dateend < tmpdate)
    		dateend = tmpdate;

    }
    else if ( set->hasBlock("RMESX1"))
    {
    	tmpdate = set->table("RMESX1")->attribute("DATE-END")->asString();
    	if (dateend < tmpdate)
    		dateend = tmpdate;

    }
    dataSetServer->close(set);
  }

  if (outSet->hasBlock("RMECX1"))
	  *outSet->table("RMECX1")->addAttribute("DATE-END") =  dateend.toString() ;
  if(outSet->hasBlock("RMESX1"))
    *outSet->table("RMESX1")->addAttribute("DATE-END") =  dateend.toString() ;

  if (outSet->hasBlock("RMECX1"))
  {
	  if (outSet->table("RMECX1")->hasColumn("NLE0"))
		  outSet->table("RMECX1")->column("NLE0")->units("c/s");
	  if (outSet->table("RMECX1")->hasColumn("NLE1"))
		  outSet->table("RMECX1")->column("NLE1")->units("c/s");
	  if (outSet->table("RMECX1")->hasColumn("NLE2"))
		  outSet->table("RMECX1")->column("NLE2")->units("c/s");
	  if (outSet->table("RMECX1")->hasColumn("NHE0"))
		  outSet->table("RMECX1")->column("NHE0")->units("c/s");
	  if (outSet->table("RMECX1")->hasColumn("NHE1"))
		  outSet->table("RMECX1")->column("NHE1")->units("c/s");
	  if (outSet->table("RMECX1")->hasColumn("NHE2"))
		  outSet->table("RMECX1")->column("NHE2")->units("c/s");
	  if (outSet->table("RMECX1")->hasColumn("NHEC"))
		  outSet->table("RMECX1")->column("NHEC")->units("c/s");
  }

}



/* Write the count rate block */
void RadmonFileMaker::writeCountRateBlock(DataSet *outSet, 
                                         std::vector<std::string> ecxfiles)
{
  // Copy input block to output file

  //Get the size of the new table

	std::map<double,std::string> sortedECXFiles;
	for (unsigned int i=0; i<ecxfiles.size(); i++) {

		DataSet *set = dataSetServer->open(ecxfiles[i],DataSet::Read);
		if (set->hasBlock("RMECX1")) {
			// Check if an output count rate (ECX) block exists in output file
			if ( !outSet->hasBlock("RMECX1")) {
				double time = set->table("RMECX1")->column("MJD")->data()->real64Data()[0];
				sortedECXFiles.insert(std::make_pair(time, ecxfiles[i]));

			}
		}
	}

	std::map<double,std::string>::iterator it = sortedECXFiles.begin();


  // Loop over input files
  Table * tab = 0;
  //for (unsigned i=0; i<ecxfiles.size(); i++) {
  while(it != sortedECXFiles.end())	{
     DataSet *set = dataSetServer->open(it->second,DataSet::Read);

     if (set->hasBlock("RMECX1")) {

       // Check if an output count rate (ECX) block exists in output file
        if (! outSet->hasBlock("RMECX1")) {

          // Copy the Block into the output file
          if (set->hasBlock("RMECX1")) {
             tab = set -> table("RMECX1");
             outSet->add(tab,"RMECX1");
          } else {
             errstr<<"Input RADMON count rate file "<<it->second
                   <<" does not contain a RMECX1 table"
                   << Sas::error("NonStandardInputFile");
          }

        } else {
         // Need to write code here to add to the arrays
        	Table *oriTab = set->table("RMECX1");
        	Table *destTab = outSet->table("RMECX1");
        	unsigned long from = destTab->rows();
        	unsigned long rows =  oriTab->rows();

        	errstr <<"Adding "<< rows << " counts rows from row " << from << Sas::message(Msg::App, Msg::Always);
        	outSet->table("RMECX1")->insertRows(from,rows);
        	Table *tab = outSet->table("RMECX1");
        	DalVecJ oriFTCoarse( oriTab->column("FTCOARSE") );
        	DalVecJ destFTCoarse( destTab->column("FTCOARSE") );
        	DalVecIu oriFTFine( oriTab->column("FTFINE") );
        	DalVecIu destFTFine( destTab->column("FTFINE") );
        	DalVecJ oriFrame( oriTab->column("FRAME") );
        	DalVecJ destFrame( destTab->column("FRAME") );
        	DalVecIu oriNLE0( oriTab->column("NLE0") );
        	DalVecIu destNLE0( destTab->column("NLE0") );
        	DalVecIu oriNLE1( oriTab->column("NLE1") );
        	DalVecIu destNLE1( destTab->column("NLE1") );
        	DalVecIu oriNLE2( oriTab->column("NLE2") );
        	DalVecIu destNLE2( destTab->column("NLE2") );
        	DalVecIu oriNHE0( oriTab->column("NHE0") );
        	DalVecIu destNHE0( destTab->column("NHE0") );
        	DalVecIu oriNHE1( oriTab->column("NHE1") );
        	DalVecIu destNHE1( destTab->column("NHE1") );
        	DalVecIu oriNHE2( oriTab->column("NHE2") );
        	DalVecIu destNHE2( destTab->column("NHE2") );
        	DalVecIu oriNHEC( oriTab->column("NHEC") );
        	DalVecIu destNHEC( destTab->column("NHEC") );
        	DalVecI oriWFSTAT( oriTab->column("WFSTAT") );
        	DalVecI destWFSTAT( destTab->column("WFSTAT") );
        	DalVecI oriEVALCADD( oriTab->column("EVALCADD") );
        	DalVecI destEVALCADD( destTab->column("EVALCADD") );
        	DalVecI oriEVALCNO( oriTab->column("EVALCNO") );
        	DalVecI destEVALCNO( destTab->column("EVALCNO") );
        	DalVecD oriTIME (oriTab->column("TIME") );
			DalVecD destTIME (destTab->column("TIME"));
			DalVecD oriMJD(oriTab->column("MJD"));
			DalVecD destMJD(destTab->column("MJD"));
			DalVecS oriTIMECAL(oriTab->column("MJD"));
			DalVecS destTIMECAL(destTab->column("MJD"));
			Column * oriCol = oriTab->column("TIMECAL");
			ColumnData * oriColData = oriCol->data();
			Column * destCol = destTab->column("TIMECAL");
			ColumnData * destColData = destCol->data();

        	unsigned long tot = from+rows;
        	for (unsigned long n = 0; n < rows; ++n)
        	{
        		destFTCoarse[from+n] = oriFTCoarse[n];
        		destFTFine[from+n]   = oriFTFine[n];
        		destFrame[from+n]    = oriFrame[n];
        		destNLE0[from+n]     = oriNLE0[n];
				destNLE1[from+n]     = oriNLE1[n];
				destNLE2[from+n]     = oriNLE2[n];
				destNHE0[from+n]     = oriNHE0[n];
				destNHE1[from+n]     = oriNHE1[n];
				destNHE2[from+n]     = oriNHE2[n];
				destNHEC[from+n]     = oriNHEC[n];
				destWFSTAT[from+n]   = oriWFSTAT[n];
				destEVALCADD[from+n] = oriEVALCADD[n];
				destEVALCNO[from+n]  = oriEVALCNO[n];
				destTIME[from+n] = oriTIME[n];
				destMJD[from+n] = oriMJD[n];
				SasDalUtils::setStringColumnRow(destColData,from+n, SasDalUtils::getStringColumnRow(oriColData,n));

        	}
        }

     }
     // Close input set
     dataSetServer->close(set);
     it++;
  }
}

/* Write the spectral block */
void RadmonFileMaker::writeSpectralBlock(DataSet *outSet, 
                                         vector<string> esxfiles)
{
  // Copy input block to output file

	std::map<double,std::string> sortedESXFiles;
	for (unsigned int i=0; i<esxfiles.size(); i++) {

		DataSet *set = dataSetServer->open(esxfiles[i],DataSet::Read);
		if (set->hasBlock("RMESX1")) {

			// Check if an output count rate (ECX) block exists in output file
			if ( !outSet->hasBlock("RMESX1")) {
				double time = set->table("RMESX1")->column("MJD")->data()->real64Data()[0];
				sortedESXFiles.insert(std::make_pair(time, esxfiles[i]));
			}
		}
	}

	std::map<double,std::string>::iterator it = sortedESXFiles.begin();


	Table * tab = 0;
    // Loop over input files

	while(it != sortedESXFiles.end())	{
     DataSet *set = dataSetServer->open(it->second,DataSet::Read);

     if (set->hasBlock("RMESX1")) {

       // Check if an output count rate (ECX) block exists in output file
        if (! outSet->hasBlock("RMESX1")) {
       
          // Copy the Block into the output file
          if (set->hasBlock("RMESX1")) {
             tab = set -> table("RMESX1");
             outSet->add(tab,"RMESX1");
          } else {
             errstr<<"Input RADMON spectral file "<<it->second
                   <<" does not contain a RMESX1 table"
                   << Sas::error("NonStandardInputFile");
          }

        } else {
       // Need to write code here to add to the arrays
        	Table *oriTab = set->table("RMESX1");

        	Table *destTab = outSet->table("RMESX1");
        	unsigned long from = destTab->rows();
        	unsigned long rows =  oriTab->rows();

        	errstr <<"Adding "<< rows << " spectra rows from row " << from << Sas::message(Msg::App, Msg::Always);
        	outSet->table("RMESX1")->insertRows(from,rows);
        	Table *tab = outSet->table("RMESX1");
        	DalVecJ oriFTCoarse( oriTab->column("FTCOARSE") );
        	DalVecJ destFTCoarse( destTab->column("FTCOARSE") );
        	DalVecIu oriFTFine( oriTab->column("FTFINE") );
        	DalVecIu destFTFine( destTab->column("FTFINE") );
        	DalVecJ oriFrame( oriTab->column("FRAME") );
        	DalVecJ destFrame( destTab->column("FRAME") );

			Column * oriCol = oriTab->column("TIMECAL");
			ColumnData * oriColData = oriCol->data();

        	DalVecD oriTIME (oriTab->column("TIME") );
			DalVecD destTIME (destTab->column("TIME"));
			DalVecD oriMJD(oriTab->column("MJD"));
			DalVecD destMJD(destTab->column("MJD"));
			DalVecS oriTIMECAL(oriTab->column("MJD"));
			DalVecS destTIMECAL(destTab->column("MJD"));

			Column * oriSPLECol = oriTab->column("SPLE");
			ColumnData * oriSPLEColData = oriSPLECol-> data();
			uint16 *oriSPLE = oriSPLEColData->uint16Data();
			std::vector<unsigned long> SPLEDimension = oriSPLEColData->size();
			Column * destSPLECol = destTab->column("SPLE");
			ColumnData * destSPLEColData = destSPLECol-> data();
			uint16 *destSPLE = destSPLEColData->uint16Data();

			Column * oriSPHESCol = oriTab->column("SPHES");
			ColumnData * oriSPHESColData = oriSPHESCol-> data();
			uint16 *oriSPHES = oriSPHESColData->uint16Data();
			Column * destSPHESCol = destTab->column("SPHES");
			ColumnData * destSPHESColData = destSPHESCol-> data();
			uint16 *destSPHES = destSPHESColData->uint16Data();

			Column * oriSPHECCol = oriTab->column("SPHEC");
			ColumnData * oriSPHECColData = oriSPHECCol-> data();
			uint16 *oriSPHEC = oriSPHECColData->uint16Data();
			Column * destSPHECCol = destTab->column("SPHEC");
			ColumnData * destSPHECColData = destSPHECCol-> data();
			uint16 *destSPHEC = destSPHECColData->uint16Data();


			Column * destCol = destTab->column("TIMECAL");
			ColumnData * destColData = destCol->data();
	       	unsigned long tot = from+rows;
	        for (unsigned long n = 0; n < rows; ++n)
	        {
	        	destFTCoarse[from+n] = oriFTCoarse[n];
	        	destFTFine[from+n]   = oriFTFine[n];
	        	destFrame[from+n]    = oriFrame[n];
	        	for (unsigned int k = 0 ; k < SPLEDimension[0] ; ++k)
	        	{
	        		destSPLE[(from*256)+(256*n)+k] = oriSPLE[(n*256)+k];
	        		destSPHES[(from*256)+(256*n)+k] = oriSPHES[(n*256)+k];
	        		destSPHEC[(from*256)+(256*n)+k] = oriSPHEC[(n*256)+k];
	        	}


	        	destTIME[from+n] = oriTIME[n];
	        	destMJD[from+n] = oriMJD[n];
	        	SasDalUtils::setStringColumnRow(destColData,from+n, SasDalUtils::getStringColumnRow(oriColData,n));

	        }


        }
     }
     // Close input set
     dataSetServer->close(set);
     it++;
  }
}

/* Destructor */
RadmonFileMaker::~RadmonFileMaker()
{
  PROC;
  ENDPROC;
}

